/*
 * Created on 23.09.2008
 *
 */
package com.eaio.twitterbackup;

import java.io.*;
import java.util.ArrayList;
import java.util.TreeSet;

import org.apache.commons.lang.text.StrBuilder;
import org.apache.log4j.Logger;

import com.eaio.twitterbackup.entities.*;
import com.thoughtworks.xstream.XStream;
import com.thoughtworks.xstream.io.json.JettisonMappedXmlDriver;
import com.thoughtworks.xstream.io.naming.NameCoder;
import com.thoughtworks.xstream.io.xml.XppDriver;

/**
 * Stores and reads tweets from the Twitter XML stream.
 * 
 * @author <a href="mailto:jb&#64;eaio.com">Johann Burkard</a>
 * @version $Id: TwitterBackup.java 7868 2016-08-27 18:09:14Z johann $
 */
public class TwitterBackup extends TreeSet<Status> {
    
    private static final Logger LOG = Logger.getLogger(TwitterBackup.class);

    public static final String VERSION = "TwitterBackup 3.2.9";

    /**
     * The serial version UID.
     */
    private static final long serialVersionUID = 8561364223228785513L;

    /**
     * Constructor for TwitterBackup.
     *
     */
    public TwitterBackup() {
        super();
    }

    /**
     * Reads tweets from a reader in XML format.
     *
     * @param r the reader
     * @return how many tweets were read
     * @throws IOException 
     */
    public int readFromXML(Reader r) throws IOException {
        int read = 0;
        Statuses s = (Statuses) createXStreamForXML().fromXML(r);
        if (s != null && s.statuses != null) {
            for (Status status : s.statuses) {
                if (add(status)) {
                    ++read;
                }
            }
        }
        return read;
    }
    
    /**
     * Reads tweets from a read in Twitter's JSON format.
     * 
     * @param r the reader
     * @return how many tweets were read
     * @throws IOException
     */
    public int readFromJSON(Reader r) throws IOException {
        int read = 0;
        StrBuilder builder = new StrBuilder();
        builder.append("{tweets:{statuses:");
        copyLarge(r, builder.asWriter(), new char[1024]);
        builder.append("}}");
        dumpTweetsIfEnabled(builder);
        Statuses s = (Statuses) createXStreamForJSON().fromXML(builder.asReader());
        if (s != null && s.statuses != null) {
            for (Status status : s.statuses) {
                if (add(status)) {
                    ++read;
                }
            }
        }
        return read;
    }

    private void dumpTweetsIfEnabled(StrBuilder builder) throws IOException {
        if (Boolean.getBoolean("dumpTweets")) {
            File tmpFile = File.createTempFile("twitterbackup-", ".txt");
            FileWriter w = null;
            try {
                w = new FileWriter(tmpFile);
                w.write(builder.toString());
                LOG.info("wrote tweets from Twitter to " + tmpFile.getAbsolutePath());
            }
            finally {
                if (w != null) {
                    try {
                        w.close();
                    }
                    catch (IOException ex) {
                        // Bla
                    }
                }
            }
        }
    }

    /**
     * Writes the internal state (tweets) to a Writer.
     *
     * @param w the writer
     */
    public void write(Writer w) {
        Statuses out = new Statuses();
        out.statuses = new ArrayList<Status>();
        out.statuses.addAll(this);
        createXStreamForXML().toXML(out, w);
    }
    
    private XStream createXStreamForXML() {
        XStream x = new XStream(new XppDriver(new NameCoder() {
            
            @Override
            public String encodeNode(String name) {
                return name;
            }
            
            @Override
            public String encodeAttribute(String name) {
                return name;
            }
            
            @Override
            public String decodeNode(String nodeName) {
                return nodeName;
            }
            
            @Override
            public String decodeAttribute(String attributeName) {
                return attributeName;
            }
            
        }));
        x.alias("statuses", Statuses.class);
        x.alias("status", Status.class);
        x.alias("user_mention", UserMention.class);
        x.addImplicitCollection(Statuses.class, "statuses");
        x.alias("user", User.class);
        x.omitField(User.class, "show_all_inline_media");
        x.aliasField("protected", User.class, "_protected");
        x.alias("url_obj", URLInfo.class);
        x.alias("hashtag", HashTag.class);
        x.alias("media", Media.class);
        x.alias("symbol", Symbol.class);
        return x;
    }

    private XStream createXStreamForJSON() {
        XStream x = new XStream(new JettisonMappedXmlDriver());
        x.setMode(XStream.NO_REFERENCES);
        x.alias("tweets", Statuses.class);
        x.addImplicitArray(Statuses.class, "statuses", Status.class);
        x.addImplicitArray(Description.class, "urls", String.class);
        
        x.addImplicitArray(StatusEntities.class, "hashtags", HashTag.class);
        x.addImplicitArray(StatusEntities.class, "media", Media.class);
        x.addImplicitArray(StatusEntities.class, "symbols", Symbol.class);
        x.addImplicitArray(StatusEntities.class, "urls", URLInfo.class);
        x.addImplicitArray(StatusEntities.class, "user_mentions", UserMention.class);
        
        x.addImplicitArray(URLs.class, "urls", URLInfo.class);
        
        x.addImplicitArray(URLInfo.class, "indices", Integer.class);
        
        x.addImplicitArray(UserMention.class, "indices", Integer.class);
        
        x.addImplicitArray(HashTag.class, "indices", Integer.class);
        
        x.addImplicitArray(Media.class, "indices", Integer.class);
        
        x.addImplicitArray(Symbol.class, "indices", Integer.class);
        
        x.addImplicitArray(ExtendedEntities.class, "media", Media.class);
        
        x.aliasField("protected", User.class, "_protected");
        return x;
    }
    
    // Via commons-io
    private static final int EOF = -1;
    
    private static long copyLarge(Reader input, Writer output, char [] buffer) throws IOException {
        long count = 0;
        int n = 0;
        while (EOF != (n = input.read(buffer))) {
            output.write(buffer, 0, n);
            count += n;
        }
        return count;
    }
    // End via
    
    /*
    public static void main(String[] args) throws IOException {
        //new TwitterBackup().readFromXML(new FileReader("johannburkard.xml"));
        TwitterBackup backup = new TwitterBackup();
        backup.readFromJSON(new FileReader("test.json"));
        backup.write(new FileWriter("rah.xml"));
    }
    */
    

}
